"""Asynchronous Python client for CEMM devices."""
from __future__ import annotations

import asyncio
import socket
from collections.abc import Mapping
from dataclasses import dataclass
from importlib import metadata
from typing import Any

import async_timeout
from aiohttp.client import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .exceptions import CEMMConnectionError, CEMMError
from .models import Connection, Device, SmartMeter, SolarPanel, WaterMeter


@dataclass
class CEMM:
    """Main class for handling connection with the CEMM device."""

    host: str
    request_timeout: float = 10.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        params: Mapping[str, str] | None = None,
    ) -> Any:
        """Handle a request to a CEMM device.

        Args:
            uri: Request URI, without '/', for example, 'status'
            method: HTTP Method to use.
            params: Extra options to improve or limit the response.

        Returns:
            A Python dictionary (text) with the response from
            the CEMM device.

        Raises:
            CEMMConnectionError: An error occurred while communicating
                with the CEMM device.
            CEMMError: Received an unexpected response from the CEMM device.
        """
        version = metadata.version(__package__)
        url = URL.build(scheme="http", host=self.host, path="/open-api/").join(URL(uri))

        headers = {
            "User-Agent": f"PythonCEMM/{version}",
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with async_timeout.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    params=params,
                    headers=headers,
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            raise CEMMConnectionError(
                "Timeout occurred while connecting to CEMM device"
            ) from exception
        except (ClientError, ClientResponseError, socket.gaierror) as exception:
            raise CEMMConnectionError(
                "Error occurred while communicating with the CEMM device"
            ) from exception

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            raise CEMMError(
                "Unexpected response from the CEMM device",
                {"Content-Type": content_type, "response": text},
            )

        return await response.json()

    async def all_connections(self) -> list[Connection]:
        """Get a list of all used aliases.

        Returns:
            A list of Connection objects.
        """
        results: list[Connection] = []

        data = await self.request("v1/io")
        for item in data["data"]:
            results.append(Connection.from_dict(item))
        return results

    async def device(self) -> Device:
        """Get the latest values from the CEMM device.

        Returns:
            A Device data object from the CEMM device API.
        """
        data = await self.request("v1")
        data = data["data"]
        return Device.from_dict(data)

    async def smartmeter(self, alias: str) -> SmartMeter:
        """Get the latest values from the CEMM device.

        Args:
            alias: the channel from which you want to read data.

        Returns:
            A SmartMeter data object from the CEMM device API.
        """
        data = await self.request(f"v1/{alias}/realtime")
        return SmartMeter.from_dict(data)

    async def watermeter(self, alias: str) -> WaterMeter:
        """Get the latest values from the CEMM device.

        Args:
            alias: the channel from which you want to
                read data (pulse-1, pulse-2, pulse-3).

        Returns:
            A WaterMeter data object from the CEMM device API.
        """
        data = await self.request(f"v1/{alias}/realtime")
        return WaterMeter.from_dict(data)

    async def solarpanel(self, alias: str) -> SolarPanel:
        """Get the latest values from the CEMM device.

        Args:
            alias: the channel from which you want to
                read data (mb1 t/m mb5).

        Returns:
            A SolarPanel data object from the CEMM device API.
        """
        data = await self.request(f"v1/{alias}/realtime")
        return SolarPanel.from_dict(data)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> CEMM:
        """Async enter.

        Returns:
            The CEMM object.
        """
        return self

    async def __aexit__(self, *_exc_info: Any) -> None:
        """Async exit.

        Args:
            _exc_info: Exec type.
        """
        await self.close()



"""Models for CEMM device."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class Connection:
    """Object representing an Connection response from CEMM."""

    io_id: int
    io_type: str
    alias: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Connection:
        """Return Connection object from the CEMM response.

        Args:
            data: The JSON data from the CEMM device.

        Returns:
            An Connection object.
        """

        return cls(
            io_id=data["io_id"],
            io_type=data["type"],
            alias=data["alias"],
        )


@dataclass
class Device:
    """Object representing an Device response from CEMM."""

    model: str
    mac: str
    version: str
    core: str

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Device:
        """Return Device object from the CEMM response.

        Args:
            data: The JSON data from the CEMM device.

        Returns:
            An Device object.
        """
        return Device(
            model=data["name"],
            mac=data["mac"],
            version=data["version"],
            core=data["core"],
        )


@dataclass
class SolarPanel:
    """Object representing an SolarPanel response from CEMM device.

    Clarification of what each value in the API stands for:
        T1: Gross energy production - low
        T2: Gross energy production - high
        T3: Device consumption - low
        T4: Device consumption - high
        Electric_energy: Net energy production - low
        Electric_energy_high: Net energy production - high
    """

    power_flow: int
    device_consumption_total: float
    device_consumption_high: float
    device_consumption_low: float

    gross_production_total: float
    gross_production_low: float
    gross_production_high: float

    net_production_total: float
    net_production_low: float
    net_production_high: float

    @staticmethod
    def from_dict(data: dict[str, Any]) -> SolarPanel:
        """Return SolarPanel object from the CEMM device response.

        Args:
            data: The JSON data from the CEMM device.

        Returns:
            An SolarPanel object.
        """

        def sum_values(value1: float, value2: float) -> float:
            total = round(float(value1 + value2), 2)
            return total

        return SolarPanel(
            power_flow=data["data"]["electric_power"][1],
            device_consumption_total=sum_values(
                data["totals"]["t3"][1], data["totals"]["t4"][1]
            ),
            device_consumption_low=data["totals"]["t3"][1],
            device_consumption_high=data["totals"]["t4"][1],
            gross_production_total=sum_values(
                data["totals"]["t1"][1], data["totals"]["t2"][1]
            ),
            gross_production_low=data["totals"]["t1"][1],
            gross_production_high=data["totals"]["t2"][1],
            net_production_total=sum_values(
                data["totals"]["electric_energy"][1],
                data["totals"]["electric_energy_high"][1],
            ),
            net_production_low=data["totals"]["electric_energy"][1],
            net_production_high=data["totals"]["electric_energy_high"][1],
        )


@dataclass
class WaterMeter:
    """Object representing an WaterMeter response from CEMM."""

    flow: float
    volume: float

    @staticmethod
    def from_dict(data: dict[str, Any]) -> WaterMeter:
        """Return Water object from the CEMM response.

        Args:
            data: The JSON data from the CEMM device.

        Returns:
            An Water object.
        """
        return WaterMeter(
            flow=data["data"]["flow"][1], volume=data["totals"]["volume"][1]
        )


@dataclass
class SmartMeter:
    """Object representing an SmartMeter response from CEMM."""

    power_flow: int | None
    gas_consumption: float | None
    energy_tariff_period: str | None

    energy_consumption_high: float
    energy_consumption_low: float
    energy_returned_high: float
    energy_returned_low: float

    billed_energy_low: float
    billed_energy_high: float

    @staticmethod
    def from_dict(data: dict[str, Any]) -> SmartMeter:
        """Return SmartMeter object from the CEMM response.

        Args:
            data: The JSON data from the CEMM device.

        Returns:
            An SmartMeter object.
        """
        return SmartMeter(
            power_flow=data["data"]["electric_power"][1],
            gas_consumption=data["data"]["gas"][1],
            energy_tariff_period=data["data"]["rate"][1],
            energy_consumption_low=data["data"]["t1"][1],
            energy_consumption_high=data["data"]["t2"][1],
            energy_returned_low=data["data"]["t3"][1],
            energy_returned_high=data["data"]["t4"][1],
            billed_energy_low=data["totals"]["electric_energy"][1],
            billed_energy_high=data["totals"]["electric_energy_high"][1],
        )



"""Asynchronous Python client for the CEMM Device."""

from .cemm import CEMM
from .exceptions import CEMMConnectionError, CEMMError
from .models import Connection, Device, SmartMeter, SolarPanel, WaterMeter

__all__ = [
    "WaterMeter",
    "Device",
    "SolarPanel",
    "SmartMeter",
    "Connection",
    "CEMM",
    "CEMMError",
    "CEMMConnectionError",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cemm']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.0.0', 'yarl>=1.6.0']

setup_kwargs = {
    'name': 'cemm',
    'version': '0.5.1',
    'description': 'Asynchronous Python client for the CEMM devices',
    'long_description': '<!-- Banner -->\n![alt Banner of the CEMM package](https://raw.githubusercontent.com/klaasnicolaas/python-cemm/main/assets/header_cemm-min.png)\n\n<!-- PROJECT SHIELDS -->\n[![GitHub Release][releases-shield]][releases]\n[![Python Versions][python-versions-shield]][pypi]\n![Project Stage][project-stage-shield]\n![Project Maintenance][maintenance-shield]\n[![License][license-shield]](LICENSE)\n\n[![GitHub Activity][commits-shield]][commits-url]\n[![PyPi Downloads][downloads-shield]][downloads-url]\n[![GitHub Last Commit][last-commit-shield]][commits-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n\n[![Code Quality][code-quality-shield]][code-quality]\n[![Build Status][build-shield]][build-url]\n[![Typing Status][typing-shield]][typing-url]\n\n[![Maintainability][maintainability-shield]][maintainability-url]\n[![Code Coverage][codecov-shield]][codecov-url]\n\nAsynchronous Python client for the CEMM devices.\n\n## About\n\nA python package with which you can read the data from your [CEMM][cemm] device via a local API. You can use it to read your smart meter via the P1, read your water meter and gain insight into how much your solar panels are producing.\n\n## Installation\n\n```bash\npip install cemm\n```\n\n## Usage\n\n```py\nimport asyncio\n\nfrom cemm import CEMM\n\n\nasync def main() -> None:\n    """Show example on getting data from your CEMM device."""\n    async with CEMM(\n        host="127.0.0.1",\n    ) as client:\n        connections = await client.all_connections()\n        device = await client.device()\n        smartmeter = await client.smartpanel("p1")\n        watermeter = await client.watermeter("pulse-1")\n        solarpanel = await client.solarpanel("mb3")\n        print(connections)\n        print(device)\n        print(smartmeter)\n        print(watermeter)\n        print(solarpanel)\n\n\nif __name__ == "__main__":\n    asyncio.run(main())\n```\n\n## Data\n\nYou can read the following data with this package, the `power flow` entities can also give a negative value.\n\n### Connections\n\n- ID\n- Type\n- Alias\n\n### Device\n\n- Model\n- Mac address\n- Version\n- Core version\n\n### SmartMeter\n\n- Power Flow (W)\n- Gas Consumption (m3)\n- Energy Tariff Period\n- Energy Consumption - High / Low (kWh)\n- Energy Returned - High / Low (kWh)\n- Billed Energy - High / Low (kWh)\n\n### WaterMeter\n\n- Flow (liters)\n- Volume\n\n### Solarpanel\n\n- Power Flow (W)\n- Device Consumption - High / Low / Total (kWh)\n- Gross Production - High / Low / Total (kWh)\n- Net Production - High / Low / Total (kWh)\n\n## Contributing\n\nThis is an active open-source project. We are always open to people who want to\nuse the code or contribute to it.\n\nWe\'ve set up a separate document for our\n[contribution guidelines](CONTRIBUTING.md).\n\nThank you for being involved! :heart_eyes:\n\n## Setting up development environment\n\nThis Python project is fully managed using the [Poetry][poetry] dependency\nmanager.\n\nYou need at least:\n\n- Python 3.9+\n- [Poetry][poetry-install]\n\nInstall all packages, including all development requirements:\n\n```bash\npoetry install\n```\n\nPoetry creates by default an virtual environment where it installs all\nnecessary pip packages, to enter or exit the venv run the following commands:\n\n```bash\npoetry shell\nexit\n```\n\nSetup the pre-commit check, you must run this inside the virtual environment:\n\n```bash\npre-commit install\n```\n\n*Now you\'re all set to get started!*\n\nAs this repository uses the [pre-commit][pre-commit] framework, all changes\nare linted and tested with each commit. You can run all checks and tests\nmanually, using the following command:\n\n```bash\npoetry run pre-commit run --all-files\n```\n\nTo run just the Python tests:\n\n```bash\npoetry run pytest\n```\n\n## License\n\nMIT License\n\nCopyright (c) 2021-2023 Klaas Schoute\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the "Software"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.\n\n[cemm]: https://cemm.nl\n\n<!-- MARKDOWN LINKS & IMAGES -->\n[build-shield]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/tests.yaml/badge.svg\n[build-url]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/tests.yaml\n[code-quality-shield]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/codeql.yaml/badge.svg\n[code-quality]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/codeql.yaml\n[commits-shield]: https://img.shields.io/github/commit-activity/y/klaasnicolaas/python-cemm.svg\n[commits-url]: https://github.com/klaasnicolaas/python-cemm/commits/main\n[codecov-shield]: https://codecov.io/gh/klaasnicolaas/python-cemm/branch/main/graph/badge.svg?token=VQTR24YFQ9\n[codecov-url]: https://codecov.io/gh/klaasnicolaas/python-cemm\n[downloads-shield]: https://img.shields.io/pypi/dm/cemm\n[downloads-url]: https://pypistats.org/packages/cemm\n[issues-shield]: https://img.shields.io/github/issues/klaasnicolaas/python-cemm.svg\n[issues-url]: https://github.com/klaasnicolaas/python-cemm/issues\n[license-shield]: https://img.shields.io/github/license/klaasnicolaas/python-cemm.svg\n[last-commit-shield]: https://img.shields.io/github/last-commit/klaasnicolaas/python-cemm.svg\n[maintenance-shield]: https://img.shields.io/maintenance/yes/2023.svg\n[maintainability-shield]: https://api.codeclimate.com/v1/badges/82ca5d035a7ef3520b52/maintainability\n[maintainability-url]: https://codeclimate.com/github/klaasnicolaas/python-cemm/maintainability\n[project-stage-shield]: https://img.shields.io/badge/project%20stage-experimental-yellow.svg\n[pypi]: https://pypi.org/project/cemm/\n[python-versions-shield]: https://img.shields.io/pypi/pyversions/cemm\n[typing-shield]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/typing.yaml/badge.svg\n[typing-url]: https://github.com/klaasnicolaas/python-cemm/actions/workflows/typing.yaml\n[releases-shield]: https://img.shields.io/github/release/klaasnicolaas/python-cemm.svg\n[releases]: https://github.com/klaasnicolaas/python-cemm/releases\n[stars-shield]: https://img.shields.io/github/stars/klaasnicolaas/python-cemm.svg\n[stars-url]: https://github.com/klaasnicolaas/python-cemm/stargazers\n\n[poetry-install]: https://python-poetry.org/docs/#installation\n[poetry]: https://python-poetry.org\n[pre-commit]: https://pre-commit.com\n',
    'author': 'Klaas Schoute',
    'author_email': 'hello@student-techlife.com',
    'maintainer': 'Klaas Schoute',
    'maintainer_email': 'hello@student-techlife.com',
    'url': 'https://github.com/klaasnicolaas/python-cemm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
